package sun.nio.ch;

import sun.security.action.GetPropertyAction;
import sun.misc.VM;
import java.lang.reflect.InvocationTargetException;
import java.nio.MappedByteBuffer;
import java.io.FileDescriptor;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.nio.ByteBuffer;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.lang.reflect.Constructor;
import sun.misc.Unsafe;

public class Util
{
    private static final int TEMP_BUF_POOL_SIZE;
    private static final long MAX_CACHED_BUFFER_SIZE;
    private static ThreadLocal<Util.BufferCache> bufferCache;
    private static Unsafe unsafe;
    private static int pageSize;
    private static volatile Constructor<?> directByteBufferConstructor;
    private static volatile Constructor<?> directByteBufferRConstructor;
    private static volatile String bugLevel;
    
    public Util() {
    }
    
    private static long getMaxCachedBufferSize() {
        String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
            @Override
            public String run() {
                return System.getProperty("jdk.nio.maxCachedBufferSize");
            }
        });
        if (s != null) {
            try {
                long long1 = Long.parseLong(s);
                if (long1 >= 0L) {
                    return long1;
                }
            }
            catch (NumberFormatException ex) {}
        }
        return Long.MAX_VALUE;
    }
    
    private static boolean isBufferTooLarge(int n) {
        return (long)n > Util.MAX_CACHED_BUFFER_SIZE;
    }
    
    private static boolean isBufferTooLarge(ByteBuffer byteBuffer) {
        return isBufferTooLarge(byteBuffer.capacity());
    }
    
    public static ByteBuffer getTemporaryDirectBuffer(int n) {
        if (isBufferTooLarge(n)) {
            return ByteBuffer.allocateDirect(n);
        }
        Util.BufferCache bufferCache = (Util.BufferCache)Util.bufferCache.get();
        ByteBuffer value = bufferCache.get(n);
        if (value != null) {
            return value;
        }
        if (!bufferCache.isEmpty()) {
            free(bufferCache.removeFirst());
        }
        return ByteBuffer.allocateDirect(n);
    }
    
    public static void releaseTemporaryDirectBuffer(ByteBuffer byteBuffer) {
        offerFirstTemporaryDirectBuffer(byteBuffer);
    }
    
    static void offerFirstTemporaryDirectBuffer(ByteBuffer byteBuffer) {
        if (isBufferTooLarge(byteBuffer)) {
            free(byteBuffer);
            return;
        }
        assert byteBuffer != null;
        if (!((Util.BufferCache)Util.bufferCache.get()).offerFirst(byteBuffer)) {
            free(byteBuffer);
        }
        return;
    }
    
    static void offerLastTemporaryDirectBuffer(ByteBuffer byteBuffer) {
        if (isBufferTooLarge(byteBuffer)) {
            free(byteBuffer);
            return;
        }
        assert byteBuffer != null;
        if (!((Util.BufferCache)Util.bufferCache.get()).offerLast(byteBuffer)) {
            free(byteBuffer);
        }
        return;
    }
    
    private static void free(ByteBuffer byteBuffer) {
        ((DirectBuffer)byteBuffer).cleaner().clean();
    }
    
    static ByteBuffer[] subsequence(ByteBuffer[] array, int n, int n2) {
        if (n == 0 && n2 == array.length) {
            return array;
        }
        ByteBuffer[] array2 = new ByteBuffer[n2];
        for (int i = 0; i < n2; ++i) {
            array2[i] = array[n + i];
        }
        return array2;
    }
    
    static <E> Set<E> ungrowableSet(final Set<E> set) {
        return new Set<E>() {
            @Override
            public int size() {
                return set.size();
            }
            
            @Override
            public boolean isEmpty() {
                return set.isEmpty();
            }
            
            @Override
            public boolean contains(Object o) {
                return set.contains(o);
            }
            
            @Override
            public Object[] toArray() {
                return set.toArray();
            }
            
            @Override
            public <T> T[] toArray(T[] array) {
                return set.toArray(array);
            }
            
            @Override
            public String toString() {
                return set.toString();
            }
            
            @Override
            public Iterator<E> iterator() {
                return set.iterator();
            }
            
            @Override
            public boolean equals(Object o) {
                return set.equals(o);
            }
            
            @Override
            public int hashCode() {
                return set.hashCode();
            }
            
            @Override
            public void clear() {
                set.clear();
            }
            
            @Override
            public boolean remove(Object o) {
                return set.remove(o);
            }
            
            @Override
            public boolean containsAll(Collection<?> collection) {
                return set.containsAll(collection);
            }
            
            @Override
            public boolean removeAll(Collection<?> collection) {
                return set.removeAll(collection);
            }
            
            @Override
            public boolean retainAll(Collection<?> collection) {
                return set.retainAll(collection);
            }
            
            @Override
            public boolean add(E e) {
                throw new UnsupportedOperationException();
            }
            
            @Override
            public boolean addAll(Collection<? extends E> collection) {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    private static byte _get(long address) {
        return Util.unsafe.getByte(address);
    }
    
    private static void _put(long address, byte x) {
        Util.unsafe.putByte(address, x);
    }
    
    static void erase(ByteBuffer byteBuffer) {
        Util.unsafe.setMemory(((DirectBuffer)byteBuffer).address(), (long)byteBuffer.capacity(), (byte)0);
    }
    
    static Unsafe unsafe() {
        return Util.unsafe;
    }
    
    static int pageSize() {
        if (Util.pageSize == -1) {
            Util.pageSize = unsafe().pageSize();
        }
        return Util.pageSize;
    }
    
    private static void initDBBConstructor() {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                try {
                    Constructor<Object> declaredConstructor = Class.forName("java.nio.DirectByteBuffer").getDeclaredConstructor(new Class[] { Integer.TYPE, Long.TYPE, FileDescriptor.class, Runnable.class });
                    declaredConstructor.setAccessible(true);
                    Util.directByteBufferConstructor = declaredConstructor;
                }
                catch (ClassNotFoundException | NoSuchMethodException | IllegalArgumentException | ClassCastException cause) {
                    throw new InternalError(cause);
                }
                return null;
            }
        });
    }
    
    static MappedByteBuffer newMappedByteBuffer(int value, long value2, FileDescriptor fileDescriptor, Runnable runnable) {
        if (Util.directByteBufferConstructor == null) {
            initDBBConstructor();
        }
        MappedByteBuffer mappedByteBuffer;
        try {
            mappedByteBuffer = (MappedByteBuffer)Util.directByteBufferConstructor.newInstance(new Object[] { new Integer(value), new Long(value2), fileDescriptor, runnable });
        }
        catch (InstantiationException | IllegalAccessException | InvocationTargetException cause) {
            throw new InternalError(cause);
        }
        return mappedByteBuffer;
    }
    
    private static void initDBBRConstructor() {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                try {
                    Constructor<Object> declaredConstructor = Class.forName("java.nio.DirectByteBufferR").getDeclaredConstructor(new Class[] { Integer.TYPE, Long.TYPE, FileDescriptor.class, Runnable.class });
                    declaredConstructor.setAccessible(true);
                    Util.directByteBufferRConstructor = declaredConstructor;
                }
                catch (ClassNotFoundException | NoSuchMethodException | IllegalArgumentException | ClassCastException cause) {
                    throw new InternalError(cause);
                }
                return null;
            }
        });
    }
    
    static MappedByteBuffer newMappedByteBufferR(int value, long value2, FileDescriptor fileDescriptor, Runnable runnable) {
        if (Util.directByteBufferRConstructor == null) {
            initDBBRConstructor();
        }
        MappedByteBuffer mappedByteBuffer;
        try {
            mappedByteBuffer = (MappedByteBuffer)Util.directByteBufferRConstructor.newInstance(new Object[] { new Integer(value), new Long(value2), fileDescriptor, runnable });
        }
        catch (InstantiationException | IllegalAccessException | InvocationTargetException cause) {
            throw new InternalError(cause);
        }
        return mappedByteBuffer;
    }
    
    static boolean atBugLevel(String anObject) {
        if (Util.bugLevel == null) {
            if (!VM.isBooted()) {
                return false;
            }
            String s = (String)AccessController.doPrivileged(new GetPropertyAction("sun.nio.ch.bugLevel"));
            Util.bugLevel = ((s != null) ? s : "");
        }
        return Util.bugLevel.equals(anObject);
    }
    
    static boolean access$100(int n) {
        return isBufferTooLarge(n);
    }
    
    static boolean access$200(ByteBuffer byteBuffer) {
        return isBufferTooLarge(byteBuffer);
    }
    
    static {
        TEMP_BUF_POOL_SIZE = IOUtil.IOV_MAX;
        MAX_CACHED_BUFFER_SIZE = getMaxCachedBufferSize();
        Util.bufferCache = new ThreadLocal<Util.BufferCache>() {
            @Override
            protected Util.BufferCache initialValue() {
                return new Util.BufferCache();
            }
        };
        Util.unsafe = Unsafe.getUnsafe();
        Util.pageSize = -1;
        Util.directByteBufferConstructor = null;
        Util.directByteBufferRConstructor = null;
        Util.bugLevel = null;
    }
    
    private static class BufferCache
    {
        private ByteBuffer[] buffers;
        private int count;
        private int start;
        
        private int next(int n) {
            return (n + 1) % Util.TEMP_BUF_POOL_SIZE;
        }
        
        BufferCache() {
            this.buffers = new ByteBuffer[Util.TEMP_BUF_POOL_SIZE];
        }
        
        ByteBuffer get(int n) {
            assert !Util.access$100(n);
            if (this.count == 0) {
                return null;
            }
            ByteBuffer[] buffers = this.buffers;
            ByteBuffer byteBuffer = buffers[this.start];
            if (byteBuffer.capacity() < n) {
                byteBuffer = null;
                int n2 = this.start;
                while ((n2 = this.next(n2)) != this.start) {
                    ByteBuffer byteBuffer2 = buffers[n2];
                    if (byteBuffer2 == null) {
                        break;
                    }
                    if (byteBuffer2.capacity() >= n) {
                        byteBuffer = byteBuffer2;
                        break;
                    }
                    continue;
                }
                if (byteBuffer == null) {
                    return null;
                }
                buffers[n2] = buffers[this.start];
            }
            buffers[this.start] = null;
            this.start = this.next(this.start);
            --this.count;
            byteBuffer.rewind();
            byteBuffer.limit(n);
            return byteBuffer;
        }
        
        boolean offerFirst(ByteBuffer byteBuffer) {
            assert !Util.access$200(byteBuffer);
            if (this.count >= Util.TEMP_BUF_POOL_SIZE) {
                return false;
            }
            this.start = (this.start + Util.TEMP_BUF_POOL_SIZE - 1) % Util.TEMP_BUF_POOL_SIZE;
            this.buffers[this.start] = byteBuffer;
            ++this.count;
            return true;
        }
        
        boolean offerLast(ByteBuffer byteBuffer) {
            assert !Util.access$200(byteBuffer);
            if (this.count >= Util.TEMP_BUF_POOL_SIZE) {
                return false;
            }
            this.buffers[(this.start + this.count) % Util.TEMP_BUF_POOL_SIZE] = byteBuffer;
            ++this.count;
            return true;
        }
        
        boolean isEmpty() {
            return this.count == 0;
        }
        
        ByteBuffer removeFirst() {
            assert this.count > 0;
            ByteBuffer byteBuffer = this.buffers[this.start];
            this.buffers[this.start] = null;
            this.start = this.next(this.start);
            --this.count;
            return byteBuffer;
        }
    }
}
